/*
 * findmem.S - Find physical memory
 *
 * Written 1998,1999 by Werner Almesberger
 */

/*
 * Non-destructive scan which restores the original values in the collect
 * phase. For this to work, we need to reduce the number of addresses visited
 * (for simplicity, we just use the result area for the backups too, so it
 * must be large enough to hold the result and the backups).
 */

/*
 * - final free list area must be dimensioned for largest possible list; we
 *   don't check for overruns
 * - STEP_SIZE must be a multiple of PAGE_SIZE
 * - the first pass must not yield more than 511 areas (we have one page (4kB),
 *   each entry occupies 8 bytes, and the last entry must be (0,0))
 */

/*
 * Register usage:
 *
 * Global (don't clobber):
 *
 * R0	pointer to page with load file page pointers
 * R8	pointer to the final free list
 *
 * findmem:
 *
 * R2	pointer to input table
 * R3	pointer to output table
 * R4	XOR pattern
 * R5	pointer to this page (input)
 *
 * scanmem, global:
 *
 * R1	data pattern
 * R5	pointer to backup area (overwritten memory words are stored here),
 *	used with pre-decrement
 * R10	copy of LR
 *
 * scanmem, mark phase:
 *
 * R6	copy of pointer to input table
 * R7	input base (we count from base+len down to base)
 * R9	current position
 *
 * scanmem, collect phase:
 *
 * R2	length of area currently being detected
 * R6	pointer to input table
 * R7	current position
 * R9	end of area
 *
 * scanmem, fixup phase:
 *
 * R7	current position
 * R9	end of area
 * R10	pointer to input table
 *
 * Global:
 *
 * R11	debugging: iobase
 * R12	debugging: sc1
 * R13	debugging: sc2
 */

/*
 * For optimum speed, put here all potential memory areas. Then set STEP_SIZE
 * to the GCD of all area sizes. The default setting is generic but slow.
 */

pmap0:	.long	map0-boot		/* pointer to map0 */

map0:
/*
 * map.inc provides the map of areas potentially containing memory and it also
 * #defines the perfect STEP_SIZE
 */
#include "map.inc"

	.long	0			/* end of map marker */
	.long	0

pat1:	.long	0xaaaaaaaa		/* 10101010... */ 
pat2:	.long	0x66666666		/* 01100110... */

findmem:
	dbgchr	#'-',r11,r12
	ldr	r2,pmap0		/* get initial guess */
	add	r2,r2,r5
	ldr	r3,[r0,#4]		/* copy first set of results into
					   scratch page */
	add	r5,r3,#PAGE_SIZE	/* use for backup too */
	ldr	r4,pat1
	bl	scanmem			/* first round */

	dbgchr	#'=',r11,r12
	ldr	r2,[r0,#4]		/* second round: copy from scratch */
	mov	r3,r8			/*   to final free list */
	add	r5,r3,#FREELIST_SIZE	/* backup area */
	ldr	r4,pat2
	bl	scanmem

#ifdef DEBUG
	dbgchr	#13,r11,r12		/* dump free list */
	dbgchr	#10,r11,r12
	mov	r3,r8
dumpfl:	ldmia	r3!,{r1,r2}
	dbgreg	#':',r1,r11,r12,r13
	dbgreg	#'+',r2,r11,r12,r13
	dbgchr	#13,r11,r12
	dbgchr	#10,r11,r12
	cmp	r1,#0			/* at end ? */
	bne	dumpfl			/* no -> next entry */
#endif

	b	bffm			/* start to whirl pages around */

/*
 * Take a list of memory areas at R2, test them with pattern R4, and store the
 * resulting list of apparently good pages at R3. In case of aliased addresses,
 * use the lowest one.
 */

scanmem:
	mov	r6,r2			/* we'll need the pointer again l8r */
	mov	r10,r2			/*   even twice ! */
smffwd:	ldmia	r2!,{r7,r9}		/* fast-forward to end of list */
	cmp	r7,#0			/* at end ? */
	bne	smffwd			/* no -> next entry */
	sub	r2,r2,#8		/* move to last non-null entry */
mkloop:	ldmdb	r2!,{r7,r9}		/* get area start and size */
	add	r9,r9,r7		/* go to end of area */
mark:	sub	r9,r9,#STEP_SIZE	/* next available area */
	dbgreg	#'M',r9,r11,r12,r13
	eor	r1,r9,r4		/* generate pattern */
	swp	r1,r1,[r9]		/* swap pattern */
	str	r1,[r5,#-4]!		/* save original content */
	dbgreg	#'^',r1,r11,r12,r13
	cmp	r7,r9			/* done with this area ? */
	bne	mark			/* no -> next step */
	cmp	r2,r6			/* was this the first entry ? */
	bne	mkloop			/* no -> mark next area */
collect:				/* collect good memory areas */
	ldmia	r6!,{r7,r9}		/* get area start and size */
	cmp	r7,#0			/* at end ? */
	beq	colfin			/* yes -> write end mark */
	add	r9,r9,r7		/* go to end of area */
colrst:	str	r7,[r3]			/* nothing accumulated so far */
	mov	r2,#0
coloop:	dbgreg	#'C',r7,r11,r12,r13
	ldr	r1,[r7]			/* get marker */
	eor	r1,r1,r4		/* apply mask */
	cmp	r1,r7			/* check pattern */
	add	r7,r7,#STEP_SIZE	/* advance pointer */
	beq	colfnd			/* found RAM: count it */
	cmp	r2,#0			/* keep current entry ? */
	addne	r3,r3,#8		/* yes -> advance pointer */
	cmp	r7,r9			/* at end ? */
	bne	colrst			/* no -> next step */
	b	collect			/* next area */
colfnd:	add	r2,r2,#STEP_SIZE	/* found RAM: increase size */
	str	r2,[r3,#4]		/* store new size */
	cmp	r7,r9			/* at end ? */
	bne	coloop			/* no -> next step */
	add	r3,r3,#8		/* keep entry */
	mov	r2,#0			/* entry is complete now */
	b	collect			/* next area */
colfin:	cmp	r2,#0			/* unfinished entry ? */
	addne	r3,r3,#8		/* yes -> advance pointer */
	str	r7,[r3],#4		/* write end mark (R7 is 0) */
	str	r7,[r3]
fixup:	ldmia	r10!,{r7,r9}		/* get area size and start */
	cmp	r7,#0			/* at end ? */
	moveq	pc,lr			/* yes -> return */
	add	r9,r9,r7		/* go to end of area */
fixlp:	ldr	r1,[r5],#4		/* get original content */
	dbgreg	#'v',r1,r11,r12,r13
	str	r1,[r7]			/* restore */
	add	r7,r7,#STEP_SIZE	/* advance pointer */
	cmp	r7,r9			/* at end ? */
	bne	fixlp			/* no -> next step */
	b	fixup			/* next area */
